using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using System.IO;

using System.Xml;

namespace GML_to_OSM
{
    public partial class GMLtoOSM : Form
    {
        public GMLtoOSM()
        {
            InitializeComponent();
        }

        private void SetGML_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofg = new OpenFileDialog();
            ofg.ShowDialog();
            SourceGML.Text = ofg.FileName;
        }

        private void LoadGML_Click(object sender, EventArgs e)
        {
            OnFile.Text = "Loading XML";
            Application.DoEvents();

            // This is very RAM hungry, Ontario needs at least a 2GB machine, and ends up eating just over 1 GB of RAM
            XmlDocument xd = new XmlDocument();
            xd.Load(SourceGML.Text);

            gridNodes.DataSource = null;
            gridSegs.DataSource = null;
            gridWays.DataSource = null;

            Nodes.Rows.Clear();
            Segs.Rows.Clear();
            Ways.Rows.Clear();

            int node_id = 1;
            int seg_id = 1;
            int way_id = 1;

            string path;
            int start;
            int end;
            bool first;

            double lon;
            double lat;

            string road_type;
            string road_number;
            string road_type_osm;

            int on_file = 1;
            int max_nodes = Int32.Parse(NodesPerFile.Text);
            int nodes_left = max_nodes;

            OnFile.Text = "1";
            Application.DoEvents();

            foreach (XmlNode res in xd.GetElementsByTagName("gml:coordinates"))
            {
                if (res.ParentNode.ParentNode.ParentNode.Name == "NRNC1:RoadSegment")
                {
                    // http://www.geobase.ca/doc/specs/pdf/Geobase_Feature_Catalogue_LRS_NRNC1.pdf
                    
                    road_type = "";
                    road_number = "";
                    road_type_osm = "";

                    foreach (XmlNode prop in res.ParentNode.ParentNode.ParentNode.ChildNodes)
                    {
                        if (prop.Name == "NRNC1:nationalRoadClass")
                        {
                            road_type = prop.InnerText;

                            switch (road_type)
                            {
                                case "Freeway": road_type_osm = "motorway"; break;
                                case "Expressway / Highway": road_type_osm = "trunk"; break;
                                case "Arterial": road_type_osm = "primary"; break;

                                case "Collector": road_type_osm = "secondary"; break;
                                case "Local / Street": road_type_osm = "tertiary"; break;
                                case "Local / Strata": road_type_osm = "tertiary"; break;

                                case "Local / Unknown": road_type_osm = "tertiary"; break;
                                case "Alleyway / Lane": road_type_osm = "unclassified"; break;
                                case "Ramp": road_type_osm = "unclassified"; break;

                                case "Resource / Recreation": road_type_osm = "unclassified"; break;
                                case "Rapid Transit": road_type_osm = "unclassified"; break;
                                case "Service Lane": road_type_osm = "unclassified"; break;

                                case "Winter": road_type_osm = "unclassified"; break;
                            }

                        }
                        else if (prop.Name == "NRNC1:routeNumber1")
                        {
                            if (prop.InnerText != "None")
                                road_number = prop.InnerText;
                        }
                        else if (prop.Name == "NRNC1:routeNameEnglish1")
                        {
                            if (prop.InnerText != "None")
                            {
                                if (road_number != "")
                                    road_number = road_number + " - " + prop.InnerText;
                                else
                                    road_number = prop.InnerText;
                            }
                        }
                    }

                    path = " " + res.InnerText + " ";

                    start = 0;
                    first = true;

                    while (start != -1)
                    {
                        start++;
                        end = path.IndexOf(",", start);
                        lat = Double.Parse(path.Substring(start, end - start));
                        start = end + 1;
                        end = path.IndexOf(" ", start);
                        lon = Double.Parse(path.Substring(start, end - start));
                        Nodes.Rows.Add(new Object[] { node_id, lon, lat });
                        
                        if (first == false)
                        {
                            Segs.Rows.Add(new Object[] { seg_id, node_id - 1, node_id });
                            Ways.Rows.Add(new Object[] { way_id, seg_id, road_number, road_type, road_type_osm });
                            seg_id++;
                        }

                        node_id++;
                        nodes_left--;

                        first = false;
                        if (end  + 1 != path.Length)
                            start = end;
                        else
                            start = -1;
                    }

                    if (nodes_left > 0)
                    {
                        way_id++;
                    }
                    else
                    {
                        if (AutoSplit.Checked)
                        {
                            SaveOSMproc("." + on_file.ToString());

                            Nodes.Rows.Clear();
                            Segs.Rows.Clear();
                            Ways.Rows.Clear();

                            node_id = 1;
                            seg_id = 1;
                            way_id = 1;
                        }

                        nodes_left = max_nodes;
                        on_file++;

                        OnFile.Text = on_file.ToString();
                        Application.DoEvents();
                    }
                }
            }

            if (AutoSplit.Checked)
            {
                SaveOSMproc("." + on_file.ToString());

                OnFile.Text = "Done";
                Application.DoEvents();

                Nodes.Rows.Clear();
                Segs.Rows.Clear();
                Ways.Rows.Clear();
            }

            if (AutoSave.Checked)
            {
                OnFile.Text = "Saving . . .";
                Application.DoEvents();

                SaveOSMproc("");

                OnFile.Text = "Done";
                Application.DoEvents();

                // This seems to be slow (hangs) so just break after "Done"
                Nodes.Rows.Clear();
                Segs.Rows.Clear();
                Ways.Rows.Clear();
            }

            OnFile.Text = "Done";
            Application.DoEvents();

            gridNodes.DataSource = Nodes;
            gridSegs.DataSource = Segs;
            gridWays.DataSource = Ways;

            MessageBox.Show("Done");
        }

        DataTable Nodes = new DataTable();
        DataTable Segs = new DataTable();
        DataTable Ways = new DataTable();
    
        private void GMLtoOSM_Load(object sender, EventArgs e)
        {
            Nodes.Columns.Add("id", typeof(Int32));
            Nodes.Columns.Add("lat", typeof(Double));
            Nodes.Columns.Add("lon", typeof(Double));

            Segs.Columns.Add("id", typeof(Int32));
            Segs.Columns.Add("from", typeof(Int32));
            Segs.Columns.Add("to", typeof(Int32));

            Ways.Columns.Add("way_id", typeof(Int32));
            Ways.Columns.Add("seg_id", typeof(Int32));
            Ways.Columns.Add("road_name", typeof(String));
            Ways.Columns.Add("type_gml", typeof(String));
            Ways.Columns.Add("type_osm", typeof(String));
        }

        private void SaveOSM_Click(object sender, EventArgs e)
        {
            SaveOSMproc("");

            MessageBox.Show("Done");
        }

        private void SaveOSMproc(string mid)
        {
            StreamWriter sw = new StreamWriter(SourceGML.Text + mid + ".osm", false);
            sw.WriteLine("<?xml version='1.0' encoding='UTF-8'?>");
            sw.WriteLine("<osm version='0.3' generator='GMLtoOSM013'>");

            foreach (DataRow row in Nodes.Rows)
            {
                sw.WriteLine("  <node id='-" + row["id"].ToString() + "' lat='" + row["lat"].ToString() + "' lon='" + row["lon"].ToString() + "' />");
            }

            int seg_offset = (int)Nodes.Rows[Nodes.Rows.Count - 1]["id"];

            foreach (DataRow row in Segs.Rows)
            {
                sw.WriteLine("  <segment id='-" + ((int)row["id"] + seg_offset).ToString() + "' from='-" + row["from"].ToString() + "' to='-" + row["to"].ToString() + "' />");
            }

            int way_offset = seg_offset + (int)Segs.Rows[Segs.Rows.Count - 1]["id"];
            DataRow last_row = Ways.NewRow();
            int last_id = 0;

            foreach (DataRow row in Ways.Rows)
            {
                if (last_id != (int)row["way_id"])
                {
                    if (last_id != 0)
                    {
                        if ((string)last_row["type_osm"] != "")
                        {
                            sw.WriteLine("    <tag k=\"highway\" v=\"" + (string)last_row["type_osm"] + "\"/>");
                        }
                        if ((string)last_row["road_name"] != "")
                        {
                            sw.WriteLine("    <tag k=\"name\" v=\"" + (string)last_row["road_name"] + "\" />");
                        }
                        sw.WriteLine("  </way>");
                    }
                    sw.WriteLine("  <way id='-" + ((int)row["way_id"] + way_offset).ToString() + "' action='modify'>");

                    last_row = row;
                    last_id = (int)row["way_id"];
                }
                sw.WriteLine("    <seg id='-" + ((int)row["seg_id"] + seg_offset).ToString() + "' />");
            }

            sw.WriteLine("  </way>");

            sw.WriteLine("</osm>");
            sw.Close();
        }
    }
}